@@ -107,6 +107,8 @@ mkHFile cf groups = unlines
     "void shChar(Char c);",
     "void shString(String s);",
     "void shIdent(String s);",
+    "void bufEscapeS(const char *s);",
+    "void bufEscapeC(const char c);",
     "void bufAppendS(const char *s);",
     "void bufAppendC(const char c);",
     "void bufReset(void);",
@@ -192,14 +194,15 @@ mkCFile cf groups = concat
       "}",
       "void ppChar(Char c, int i)",
       "{",
-      "  char tmp[4];",
-      "  sprintf(tmp, \"'%c'\", c);",
-      "  renderS(tmp);",
+      "  bufAppendC('\\'');",
+      "  bufEscapeC(c);",
+      "  bufAppendC('\\'');",
+      "  bufAppendC(' ');",
       "}",
       "void ppString(String s, int i)",
       "{",
       "  bufAppendC('\\\"');",
-      "  bufAppendS(s);",
+      "  bufEscapeS(s);",
       "  bufAppendC('\\\"');",
       "  bufAppendC(' ');",
       "}",
@@ -235,13 +238,13 @@ mkCFile cf groups = concat
       "void shChar(Char c)",
       "{",
       "  bufAppendC('\\'');",
-      "  bufAppendC(c);",
+      "  bufEscapeC(c);",
       "  bufAppendC('\\'');",
       "}",
       "void shString(String s)",
       "{",
       "  bufAppendC('\\\"');",
-      "  bufAppendS(s);",
+      "  bufEscapeS(s);",
       "  bufAppendC('\\\"');",
       "}",
       "void shIdent(String s)",
@@ -257,7 +260,7 @@ mkCFile cf groups = concat
       "void sh" ++ t ++ "(String s)",
       "{",
       "  bufAppendC('\\\"');",
-      "  bufAppendS(s);",
+      "  bufEscapeS(s);",
       "  bufAppendC('\\\"');",
       "}",
       ""
@@ -265,6 +268,26 @@ mkCFile cf groups = concat
      ]
     footer = unlines
      [
+      "void bufEscapeS(const char *s)",
+      "{",
+      "  if (s) while (*s) bufEscapeC(*s++);",
+      "}",
+      "void bufEscapeC(const char c)",
+      "{",
+      "  switch(c)",
+      "  {",
+      "    case '\\f': bufAppendS(\"\\\\f\" ); break;",
+      "    case '\\n': bufAppendS(\"\\\\n\" ); break;",
+      "    case '\\r': bufAppendS(\"\\\\r\" ); break;",
+      "    case '\\t': bufAppendS(\"\\\\t\" ); break;",
+      "    case '\\v': bufAppendS(\"\\\\v\" ); break;",
+      "    case '\\\\': bufAppendS(\"\\\\\\\\\"); break;",
+      "    case '\\'': bufAppendS(\"\\\\'\" ); break;",
+      "    case '\\\"': bufAppendS(\"\\\\\\\"\"); break;",
+      "    default: bufAppendC(c);",
+      "  }",
+      "}",
+      "",
       "void bufAppendS(const char *s)",
       "{",
       "  size_t len = strlen(s);",
