@@ -2,42 +2,39 @@
     BNF Converter: XML generator
     Copyright (C) 2004  Author:  Aarne Ranta
 
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
 -}
 
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE PatternSynonyms #-}
+
+{-# OPTIONS_GHC -fno-warn-missing-signatures #-}
+
 module BNFC.Backend.XML ---- (cf2DTD, cf2XML)
   where
 
+import Prelude hiding ((<>))
+
+import Data.Bifunctor ( second )
+import Data.List      ( intercalate )
+
 import BNFC.CF
+import BNFC.PrettyPrint
 import BNFC.Utils
-import BNFC.Backend.Base
-import BNFC.Options hiding (Backend)
+import BNFC.Backend.Base (Backend, mkfile)
+import BNFC.Options (SharedOptions(..), pattern TargetHaskellGadt)
 import BNFC.Backend.Haskell.CFtoTemplate ()
-import BNFC.Backend.Haskell.HsOpts (xmlFile, xmlFileM, absFileM)
-import Data.List (intersperse, intercalate)
-import Data.Char(toLower)
-import Data.Maybe (fromJust)
+import BNFC.Backend.Haskell.HsOpts ( xmlFile, xmlFileM, absFileM )
+import BNFC.Backend.Haskell.Utils  ( catToVar )
+import qualified BNFC.Backend.Haskell.Utils as Haskell
 
 type Coding = Bool ---- change to at least three values
 
 makeXML :: SharedOptions -> Coding -> CF -> Backend
 makeXML opts typ cf = do
   let name = lang opts
-  mkfile (name ++ ".dtd") $ cf2DTD typ name cf
+  mkfile (name ++ ".dtd") comment $ cf2DTD typ name cf
   let absmod = "XML" ++ name
-  mkfile (xmlFile opts) $ cf2XMLPrinter typ opts absmod cf
+  mkfile (xmlFile opts) Haskell.comment $ cf2XMLPrinter typ opts absmod cf
 
 -- derive a DTD from a BNF grammar. AR 21/8/2004
 cf2DTD :: Coding -> String -> CF -> String
@@ -53,6 +50,9 @@ cf2DTD typ name cf = unlines [
   "]>"
   ]
 
+comment :: String -> String
+comment x = unwords [ "<!--", x, "-->" ]
+
 -- | >>> tag "test"
 -- "<test>"
 tag :: String -> String
@@ -68,15 +68,13 @@ elemAtt t a ts = element t ts ++++ attlist t a
 elemt t = elemAtt t "name"
 
 elemc :: Cat -> [(Fun, String)] -> String
-elemc cat fs = unlines $ element (show cat) (map snd fs) : [element f [] | (f,_) <- fs]
+elemc cat fs = unlines $ element (prettyShow cat) (map snd fs) : [element f [] | (f,_) <- fs]
 
 elemEmp :: String -> String
 elemEmp t = elemAtt t "value" []
 
 alts :: [String] -> String
-alts ts =
-  if null ts then "EMPTY" else parenth (unwords (intersperse "|" ts))
-
+alts ts = if null ts then "EMPTY" else parenth $ intercalate " | " ts
 
 -- choose between these two encodings:
 
@@ -89,24 +87,28 @@ endtagDef b = if b then endtagDefConstr else endtagDefNotyp
 -- lengthy, but validation guarantees type correctness
 -- flag -xmlt
 elemDataConstrs cf (cat,fcs) = elemc cat [(f,rhsCat cf f cs) | (f,cs) <- fcs]
-efunDefConstrs = "elemFun i t x = [replicate (i+i) ' ' ++ tag t ++ \" \" ++ etag x]"
-endtagDefConstrs = "endtag f c = tag (\"/\" ++ c)"
+
+efunDefConstrs :: String
+efunDefConstrs = "elemFun i t x = [P.replicate (i+i) ' ' ++ tag t ++ \" \" ++ etag x]"
+
+endtagDefConstrs :: String
+endtagDefConstrs = "endtag _ c = tag (\"/\" ++ c)"
 
 -- coding 1:
 -- to show constructors as empty tags;
 -- shorter than 0, but validation still guarantees type correctness
 -- flag -xmlt
 elemDataConstr cf (cat,fcs) = elemc cat [(f,rhsCat cf f cs) | (f,cs) <- fcs]
-efunDefConstr = "elemFun i t x = [replicate (i+i) ' ' ++ tag t ++ \" \" ++ etag x]"
-endtagDefConstr = "endtag f c = tag (\"/\" ++ c)"
+efunDefConstr = "elemFun i t x = [P.replicate (i+i) ' ' ++ tag t ++ \" \" ++ etag x]"
+endtagDefConstr = "endtag _ c = tag (\"/\" ++ c)"
 
 -- coding 2:
 -- constructors as tags, no types.
 -- clumsy DTD, but nice trees. Validation guarantees type correctness
 -- flag -xml
 elemDataNotyp cf (_,fcs) = unlines [element f [rhsCatNot cf cs] | (f,cs) <- fcs]
-efunDefNotyp = "elemFun i t x = [replicate (i+i) ' ' ++ tag x]"
-endtagDefNotyp = "endtag f c = tag (\"/\" ++ f)"
+efunDefNotyp = "elemFun i _ x = [P.replicate (i+i) ' ' ++ tag x]"
+endtagDefNotyp = "endtag f _ = tag (\"/\" ++ f)"
 
 
 -- to show constructors as attributes;
@@ -116,27 +118,31 @@ endtagDefNotyp = "endtag f c = tag (\"/\" ++ f)"
 -- efunDefAttr =  "elemFun i t x = [replicate (i+i) ' ' ++ tag (t ++ \" name = \" ++ x)]"
 
 rhsCat :: CF -> Fun -> [Cat] -> String
-rhsCat cf fun cs = parenth (intercalate ", " (fun:map (symbCat cf) cs))
-rhsCatNot cf cs = if null cs then "EMPTY" else intercalate", " (map (symbCatNot cf) cs)
+rhsCat cf fun cs = parenth (intercalate ", " (fun : map (render . symbCat cf) cs))
+rhsCatNot cf cs = if null cs then "EMPTY" else intercalate ", " (map (render . symbCatNot cf) cs)
 
+symbCat :: CF -> Cat -> Doc
 symbCat cf c
-  | isList c  = show (normCatOfList c) ++ if isEmptyListCat cf c then "*" else "+"
-  | otherwise = show c
+  | isList c  = pretty (normCatOfList c) <> if isEmptyListCat cf c then "*" else "+"
+  | otherwise = pretty c
 
+symbCatNot :: CF -> Cat -> Doc
 symbCatNot cf c
-  | isList c  = funs (normCatOfList c) ++ if isEmptyListCat cf c then "*" else "+"
+  | isList c  = funs (normCatOfList c) <> if isEmptyListCat cf c then "*" else "+"
   | otherwise = funs c
  where
-   funs k = case lookup k (cf2data cf) of
-     Just []  -> "EMPTY"
-     Just fcs -> parenth $ unwords $ intersperse "|" $ map fst fcs
-     _ -> parenth (show k) ----
+   funs k =
+     case lookup k (cf2data cf) of
+       Just []  -> "EMPTY"
+       Just fcs -> parens $ sep $ punctuate "|" $ map (text . fst) fcs
+       _        -> parens $ pretty k
 
 parenth s = "(" ++ s ++ ")"
 
 -- derive an XML printer from a BNF grammar
 cf2XMLPrinter :: Bool -> SharedOptions -> String -> CF -> String
 cf2XMLPrinter typ opts absMod cf = unlines [
+  "{-# LANGUAGE LambdaCase #-}",
   pragmas opts,
   prologue typ opts absMod,
   integerRule cf,
@@ -150,36 +156,54 @@ cf2XMLPrinter typ opts absMod cf = unlines [
 pragmas :: SharedOptions -> String
 pragmas opts =
   if target opts == TargetHaskellGadt
-  then "{-# LANGUAGE TypeSynonymInstances, FlexibleInstances, GADTs #-}"
+  then "{-# LANGUAGE FlexibleInstances, GADTs #-}"
   else ""
 
 prologue :: Bool -> SharedOptions -> String -> String
 prologue b opts _ = unlines [
-  "module " ++ xmlFileM opts +++ "where\n",
-  "-- pretty-printer generated by the BNF converter\n",
+  "-- Pretty printing to XML",
+  "",
+  "module " ++ xmlFileM opts +++ "where",
+  "",
+  "import Prelude",
+  "  ( Char, Double, Integer, String",
+  "  , (.), ($), (+), (++)",
+  "  )",
+  "import qualified Prelude as P",
+  "  ( Show(..), Int",
+  "  , concat, concatMap, replicate, unlines",
+  "  )",
+  "",
   "import " ++ absFileM opts,
-  "import Data.Char",
   "",
   "-- the top-level printing method",
   "printXML :: XPrint a => a -> String",
   "printXML = render . prt 0",
   "",
-  "render = unlines",
+  "render :: [String] -> String",
+  "render = P.unlines",
   "",
   "-- the printer class does the job",
   "class XPrint a where",
-  "  prt :: Int -> a -> [String]",
-  "  prtList :: Int -> [a] -> [String]",
-  "  prtList i = concat . map (prt i)",
+  "  prt :: P.Int -> a -> [String]",
+  "  prtList :: P.Int -> [a] -> [String]",
+  "  prtList = P.concatMap . prt",
   "",
   "instance XPrint a => XPrint [a] where",
   "  prt = prtList",
   "",
+  "tag, etag :: String -> String",
   "tag t = \"<\" ++ t ++ \">\"",
   "etag t = \"<\" ++ t ++ \"/>\"",
-  "elemTok i t x = [replicate (i+i) ' ' ++ tag (t ++ \" value = \" ++ show x ++ \" /\")]",
-  "elemTokS i t x = elemTok i t (show x)",
+  "",
+  "elemTok, elemTokS :: P.Show a => P.Int -> String -> a -> [String]",
+  "elemTok i t x = [P.replicate (i+i) ' ' ++ tag (t ++ \" value = \" ++ P.show x ++ \" /\")]",
+  "elemTokS i t x = elemTok i t (P.show x)",
+  "",
+  "elemFun :: P.Int -> String -> String -> [String]",
   efunDef b,
+  "",
+  "endtag :: String -> String -> String",
   endtagDef b,
   ""
   ]
@@ -205,36 +229,24 @@ ownPrintRule cf cat = unlines $
 
 rules :: CF -> String
 rules cf = unlines $
-  map (\(s,xs) -> case_fun s (map toArgs xs)) $ cf2data cf
+  map (\ (s, xs) -> case_fun s (map (second toArgs) xs)) $ cf2data cf
  where
-   toArgs (cons,args) = ((cons, names (map (checkRes . var) args) (0 :: Int)), ruleOf cons)
+   toArgs args = names (map (catToVar ["prt"]) args) (0 :: Int)
    names [] _ = []
    names (x:xs) n
      | x `elem` xs = (x ++ show n) : names xs (n+1)
-     | otherwise = x             : names xs n
-   var (ListCat c)  = var c ++ "s"
-   var (Cat "Ident")   = "id"
-   var (Cat "Integer") = "n"
-   var (Cat "String")  = "str"
-   var (Cat "Char")    = "c"
-   var (Cat "Double")  = "d"
-   var cat            = map toLower (show cat)
-   checkRes s
-        | s `elem` reservedHaskell = s ++ "'"
-        | otherwise              = s
-   reservedHaskell = ["case","class","data","default","deriving","do","else","if",
-                          "import","in","infix","infixl","infixr","instance","let","module",
-                          "newtype","of","then","type","where","as","qualified","hiding"]
-   ruleOf s = fromJust $ lookupRule s (cfgRules cf)
+     | otherwise   = x             : names xs n
 
--- case_fun :: Cat -> [(Constructor,Rule)] -> String
-case_fun cat xs = unlines [
-  "instance XPrint" +++ show cat +++ "where",
-  "  prt i" +++ "e = case e of",
-  unlines $ map (\ ((c,xx),_) ->
-    "   " ++ c +++ unwords xx +++ "-> concat $ " +++
-    "elemFun i \"" ++ show cat ++ "\" \"" ++ c ++ "\"" +++
-    unwords [": prt (i+1)" +++ x | x <- xx] +++ ":" +++
-    "[[replicate (i+i) ' ' ++ endtag \"" ++ c ++ "\" \"" ++ show cat ++ "\"]]"
-    ) xs
+case_fun :: Cat -> [(String, [String])] -> String
+case_fun cat xs = unlines $ concat
+  [ [ "instance XPrint" +++ s +++ "where"
+    , "  prt i'" +++ "= \\case"
+    ]
+  , (`map` xs) $ \ (c, xx) ->
+    "   " ++ c +++ unwords xx +++ "-> P.concat $ " +++
+    "elemFun i' \"" ++ s ++ "\" \"" ++ c ++ "\"" +++
+    unwords [": prt (i'+1)" +++ x | x <- xx] +++ ":" +++
+    "[[P.replicate (i'+i') ' ' ++ endtag \"" ++ c ++ "\" \"" ++ s ++ "\"]]"
   ]
+  where
+  s = prettyShow cat
