@@ -13,6 +13,7 @@ module BNFC.Options
   , defaultOptions, isDefault, printOptions
   , AlexVersion(..), HappyMode(..), OCamlParser(..), JavaLexerParser(..)
   , RecordPositions(..), TokenText(..)
+  , Ansi(..)
   , InPackage
   , removedIn290
   , translateOldOptions
@@ -97,9 +98,14 @@ data OCamlParser = OCamlYacc | Menhir
 data JavaLexerParser = JLexCup | JFlexCup | Antlr4
     deriving (Eq,Show,Ord)
 
+-- | Line numbers or not?
 data RecordPositions = RecordPositions | NoRecordPositions
     deriving (Eq,Show,Ord)
 
+-- | Restrict to ANSI standard (C/C++)?
+data Ansi = Ansi | BeyondAnsi
+    deriving (Eq, Ord, Show)
+
 -- | Package name (C++ and Java backends).
 type InPackage = Maybe String
 
@@ -119,9 +125,10 @@ data SharedOptions = Options
   , outDir      :: FilePath        -- ^ Target directory for generated files.
   , force       :: Bool            -- ^ Ignore errors as much as possible?
   , target      :: Target          -- ^ E.g. @--haskell@.
-  , make        :: Maybe String    -- ^ The name of the Makefile to generate or Nothing for no Makefile.
+  , optMake     :: Maybe String    -- ^ The name of the Makefile to generate or Nothing for no Makefile.
   , inPackage   :: InPackage       -- ^ The hierarchical package to put the modules in, or Nothing.
   , linenumbers :: RecordPositions -- ^ Add and set line_number field for syntax classes
+  , ansi        :: Ansi            -- ^ Restrict to the ANSI language standard (C/C++)?
   --- Haskell specific:
   , inDir         :: Bool        -- ^ Option @-d@.
   , functor       :: Bool        -- ^ Option @--functor@.  Make AST functorial?
@@ -152,9 +159,10 @@ defaultOptions = Options
   , outDir          = "."
   , force           = False
   , target          = TargetHaskell
-  , make            = Nothing
+  , optMake         = Nothing
   , inPackage       = Nothing
   , linenumbers     = NoRecordPositions
+  , ansi            = BeyondAnsi
   -- Haskell specific
   , inDir           = False
   , functor         = False
@@ -208,9 +216,10 @@ printOptions opts = unwords . concat $
   [ [ printTargetOption tgt ]
   -- General and shared options:
   , unlessDefault outDir opts $ \ o -> [ "--outputdir=" ++ o ]
-  , [ "--makefile=" ++ m  | m <- maybeToList $ make opts        ]
+  , [ "--makefile=" ++ m  | m <- maybeToList $ optMake opts        ]
   , [ "-p " ++ p          | p <- maybeToList $ inPackage opts   ]
   , unlessDefault linenumbers opts $ const [ "-l" ]
+  , unlessDefault ansi opts $ const [ "--ansi" ]
   -- Haskell options:
   , [ "-d"                | inDir opts                          ]
   , [ "--functor"         | functor opts                        ]
@@ -313,6 +322,10 @@ specificOptions =
         , "(Note: Java requires cup version 0.11b-2014-06-11 or greater.)"
         ]
     , [TargetC, TargetCpp, TargetJava] )
+  , ( Option [] ["ansi"] (NoArg (\o -> o{ ansi = Ansi })) $ unlines
+        [ "Restrict to ANSI language standard"
+        ]
+    , [TargetCpp] )  -- In the future maybe also: TargetC
   , ( Option ['p'] ["name-space"]
       (ReqArg (\n o -> o {inPackage = Just n}) "NAMESPACE")
           "Prepend NAMESPACE to the package/module name"
@@ -387,7 +400,7 @@ commonOptions =
   , Option ""  ["force"]     (NoArg (\ o -> o { force = True }))
       "Ignore errors in the grammar (may produce ill-formed output or crash)"
   ]
-  where setMakefile mf o = o { make = Just mf }
+  where setMakefile mf o = o { optMake = Just mf }
 
 allOptions :: [OptDescr (SharedOptions -> SharedOptions)]
 allOptions = targetOptions ++ commonOptions ++ map fst specificOptions
